<?php
 $g_bbparser = new BBParser();

 // Available tag names.
 // Adding a new tag here isn't enough. It must also be handled in the
 // bbtag->render() function.
 $bb_tagnames=array("b","i","u","s","f","color","url","img","email","code","center","large","small","tag","tab","sub","quote","table","tr","td","ul","li");

 class BBParser
 {
  
  function parse($string)
  {
   $tagarray = array();

   // Just padding this tag at the end to make the code for parsing a
   // little less messy.  Don't want to add special code for extracting
   // text at the end of parsing.
   // The pad tag is removed from the tag array before a tag tree is built
   $string.="[END]";

   // First read the whole string and build a tag array
   $id = 0;
   $mode = 0;
   $tagstart = 0;
   $tagend = 0;
   for($i=0;$i<strlen($string);$i++)
   {
    $c = $string[$i];
 
    switch($mode)
    {

     case(0): // reading
              if($c=="[")
              {
               $mode = 1;
               $tagstart = $i+1;
              }
              break;

     case(1): // looking for possible tag
              if($c=="[")
              {
               $tagstart++;
              }
              if($c=="\n")
              {
               $mode = 0; // not a tag, continue
              }
              else if($c=="]")
              {
               $tag = new bbtag();
               $tag->type = "TEXT";
               $tag->text = substr($string,$tagend,$tagstart-$tagend-1);
               if($tag->text!="")
               {
                $tag->id = $id;
                $id++;
                array_push($tagarray,$tag);
               }

               $tag = new bbtag();
               $tag->type = "NODE";
               $tag->text = substr($string,$tagstart,$i-$tagstart);
               $tag->validate();
               $tag->id = $id;
               $id++;
               array_push($tagarray,$tag);

               // special handling of [tab] tags. Need to add an extra end tag
               // to the list or the tree parser will go nuts
               if($tag->name=="tab")
               {
                $tag2 = new bbtag();
                $tag2->type = "NODE";
                $tag2->text = "/".$tag->name;
                $tag2->validate();
                $tag2->id = $id;
                $id++;
                array_push($tagarray,$tag2);
               }
              
               $tagend = $i+1;
               $mode = 0;
              }
              break;
    }
   }

   // Remove unused [END] tag
   array_pop($tagarray);

   // Invalidate all tags inside [code] blocks
   // no node tags are allowed in code segments
   $iscode = false;
   $tagarray2 = array();
   foreach($tagarray AS $id => $tag)
   {
    if($tag->type=="NODE" && $tag->name=="code")
    {
     array_push($tagarray2,$tag);
     $iscode = true;
    }
    else if($tag->type=="NODE" && $tag->name=="/code")
    {
     array_push($tagarray2,$tag);
     $iscode = false;
    }
    else
    {
     if($iscode && $tag->type=="NODE")
     {
      $tag->invalidate();
     }
     array_push($tagarray2,$tag);
    }
   }

   $tagarray = $tagarray2;

   // then reshuffle it all into a node tree of tags
   $root = new bbtag();
   $root->type = "NODE";
   $root->name = "";
   $root->text = "";
   $parent = $root;
   foreach($tagarray AS $id => $tag)
   {
    if($tag->type=="NODE")
    {
     $isend = $tag->name[0]=="/";
     if($isend)
     {
      if("/$parent->name"==$tag->name)
      {
       // close parent node and move back into the prevous parent
       // see to that the close tag text is stored in parent
       $parent->closeText = $tag->text;
       $parent = $parent->parent;
      }
      else
      {
       $tag->invalidate();
       $parent->addChild($tag);
      }
     }
     else
     {
      $tag->parent = $parent;
      $parent->addChild($tag);
      $parent = $tag;
     }
    }
    else
    {
     $tag->parent = $parent;
     $parent->addChild($tag);
    }
   }

   return $root->render();
  }

 }

 class bbtag
 {
  var $parent;
  var $type;
  var $name;
  var $text;
  var $params;
  var $paramList;
  var $children;
  var $suppress;
  var $id;
  var $closeText;
  var $ctr;
 
  function bbtag()
  {
   $this->parent="";
   $this->paramList = array();
   $this->children = array();
   $this->suppress = false;
  }

  // Converts a node into a text
  function invalidate()
  {
   $this->type="TEXT";
   $this->text="[".$this->text."]";
  }

  // Adds a child node into this nodes child node list
  function addChild($child)
  {
   if(!is_array($this->children))
   {
    $this->children = array();
   }

   array_push($this->children,$child);
  }

  // initital checking of a node and parameter extraction
  // Nodes can be of type:
  // [name]
  // [name=value]
  // [name value1=a ...]
  // If the node is not a valid name or isn't of any of the above
  // types it is converted into a TEXT node

  function validate()
  {
   global $bb_tagnames;

   if($this->type=="NODE") 
   {
    $data = explode(" ",$this->text);
    $err = true;
    if(count($data)>0)
    {
     $type = $data[0];

     if($type!="")
     {
      if($type[0]=="/")
      {
       $type = substr($type,1);
      }
     }

     if(in_array($type,$bb_tagnames))
     {
      $this->name = $data[0];
      if($type!=$this->text && $data[0][0]!="/")
      {
       $this->params = substr($this->text,strlen($type)+1);
       $this->extractParams();
      }
      $err = false;
     }
    }

    if($err)
    {
     $str = $this->text;
     $data = explode(" ",$str);
     if(count($data)!=0)
     {
      $str = $data[0];
     }

     $data = explode("=",$str);
     if(count($data)>0)
     {
      $type = $data[0];
      if(in_array($type,$bb_tagnames))
      {
       $this->name = $type;
       $this->params = $this->text;
       $this->extractParams();
       $err = false;
      }
     }
    }

    if($err)
    {
     $this->type = "TEXT";
     $this->text = "[".$this->text."]";
    }

   }
  }

  // If you want an overview of the node tree then this function will
  // print it for you.
  function printTree($dis="")
  {
   echo "$dis NODE ".$this->toString()."<BR />";
   if(is_array($this->children))
   {
    foreach($this->children AS $id =>$child)
    {
     if($child->type=="NODE")
     {
      $child->printTree("$dis-");
     }
     else
     {
      echo "-$dis TEXT ".$child->toString()."<BR />";
     }
    }
   }
  }

  function render($puretext=false)
  {
   global $col_tableframe;
   global $col_quotebg,$col_codebg,$col_mainbg;
   global $col_tablebg1,$col_tablebg2;
   global $col_tableheaderfg,$col_tableheaderbg;

   $result = "";
   if($puretext)
   {
    if($this->type=="TEXT")
    {
     $result.= $this->text;
    }
    else if($this->type=="NODE")
    {
     $result.= "[".$this->text."]";
     $children = $this->children;
     foreach($children AS $id => $child)
     {
      $result.= $child->render(true);
     }
     if($this->closeText!="")
     {
      $result.= "[".$this->closeText."]";
     }
     else
     {
      $result.= "[/".$this->text."]";
     }
    }
   }
   else if($this->type=="TEXT")
   {
    $result.= str_replace("\n","<BR />",$this->text);
   }
   else if($this->type=="NODE")
   {
    $dochildren = true;
    $dotext = true;
    switch(strtolower($this->name))
    {
     case "b": $start = "<STRONG>";
               $end = "</STRONG>";
               break;
     case "i": $start = "<EM>";
               $end = "</EM>";
               break;
     case "u": $start = "<U>";
               $end = "</U>";
               break;
     case "s": $start = "<STRIKE>";
               $end = "</STRIKE>";
               break;
     case "f": $start = "<CODE>";
               $end = "</CODE>";
               break;
     case "color":
               $color = $this->paramList["color"];
               // Should validate color here
               if($color!="")
               {
                $start = "<FONT COLOR='$color'>";
                $end = "</FONT>";
               }
               else
               {
                $start = "[$this->text]";
                $end = "[/color]";
               }
               break;
     case "url":
               $url = $this->paramList["url"];
               if($url!="")
               {
                $start = "<A HREF='$url'>";
                $end = "</A>";
               }
               else
               {
                $dochildren = false;
                $children = $this->children;
                $url = "";
                foreach($children AS $id => $child)
                {
                 $url.= $child->render();
                }
                $result.="<A HREF='$url'>$url</A>";
                $start = "";
                $end = "";
               }
               break;
     case "img":
               $url = $this->paramList["img"];
               if($url!="")
               {
                $start = "<A HREF='$url'><IMG BORDER='0' ALT='' SRC='";
                $end = "'></IMG></A>";
               }
               else
               {
                $start = "<IMG ALT='' SRC='";
                $end = "'></IMG>";
               }
               break;   
     case "email":
                $dochildren = false;
                $children = $this->children;
                $email = "";
                foreach($children AS $id => $child)
                {
                 $email.= $child->render();
                }
                $result.="<A HREF='mailto:$email'>$email</A>";
                $start = "";
                $end = "";
                break;
     case "code":
                $dochildren = false;
                $children = $this->children;
                $code = "";
                foreach($children AS $id => $child)
                {
                 $code.= $child->render(true);
                }
                $start = "<TABLE BORDER='0' CELLPADDING='0' CELLSPACING='1' WIDTH='100%'><TR BGCOLOR='$col_tableframe'><TD><TABLE BORDER='0' CELLSPACING='1' CELLPADDING='4' WIDTH='100%'><TR><TD BGCOLOR='$col_codebg'><PRE>$code</PRE></TD></TR></TABLE></TD></TR></TABLE>";
                $end = "";
                break;
     case "center":
                $start = "<CENTER>";
                $end = "</CENTER>";
                break;
     case "large":
                $large = $this->paramList["large"];
                // validate color
                if($large!="")
                {
                 $start = "<FONT SIZE='+1' COLOR='$large'>";
                 $end = "</FONT>";
                }
                else
                {
                 $start = "<FONT SIZE='+1'>";
                 $end = "</FONT>";
                }
                break;
     case "small":
                $small = $this->paramList["small"];
                // validate color
                if($small!="")
                {
                 $start = "<FONT SIZE='-2' COLOR='$small'>";
                 $end = "</FONT>";
                }
                else
                {
                 $start = "<FONT SIZE='-2'>";
                 $end = "</FONT>";
                }
                break;
     case "tag":
                $start = "<A ID='";
                $end = "'></A>";
                break;
     case "sub":
                $start = "<TABLE WIDTH='100%' BORDER='0' CELLPADDING='0' CELLSPACING='0'><TR><TD NOWRAP>&nbsp;&nbsp;&nbsp;&nbsp;</TD><TD WIDTH='100%'>";
                $end = "</TD></TR></TABLE>";
                break;
     case "quote":
                $start = "<TABLE BORDER='0' CELLPADDING='0' CELLSPACING='1' WIDTH='100%'><TR BGCOLOR='$col_tableframe'><TD><TABLE BORDER='0' CELLSPACING='1' CELLPADDING='4' WIDTH='100%'><TR><TD BGCOLOR='$col_quotebg'>";
                $end = "</TD></TR></TABLE></TD></TR></TABLE>";
                break;
     case "tab":
                $start = "&nbsp;&nbsp;&nbsp;&nbsp;";
                $end = "";
                break;
     case "tr":
                $children = $this->children;
                $hadnodes = false;
                foreach($children AS $id => $child)
                {
                 if($child->type=="NODE" && $child->name=="td")
                 {
                  $hasnodes = true;
                 }
                }

                $table = $this->parent;

                if($hasnodes && is_object($table) && $table->type=="NODE" && $table->name=="table" && !$table->suppress)
                {
                 if($table->paramList["header"]=="on" && $table->ctr == 0)
                 {
                  $bgcolor = $col_tableheaderbg;
                 }
                 else if($table->paramList["stripes"]=="on")
                 {
                  $lctr = $table->ctr;
                  if($table->paramList["header"]=="on")
                  {
                   $lctr++;
                  }
                  if($lctr%2==0)
                  {
                   $bgcolor = $col_tablebg1;
                  }
                  else
                  {
                   $bgcolor = $col_tablebg2;
                  }
                 }
                 else
                 {
                  $bgcolor = $col_mainbg;
                 }
                 $start = "<TR BGCOLOR='$bgcolor'>";
                 $end = "</TR>";
                 $dotext = false;

                 $table->ctr = $table->ctr+1;
                }
                else
                {
                 $start = "[".$this->text."]";
                 $end = "[/tr]";
                }
                break;
     case "td":
                $row = $this->parent;
                $table = "";
                if(is_object($row))
                {
                 $table = $row->parent;
                }
                if(is_object($row) && $row->type=="NODE" && $row->name=="tr" && is_object($table) && $table->type=="NODE" && $table->name=="table" && !$table->suppress)
                {
                 if($table->paramList["header"]=="on" && $table->ctr == 1)
                 {
                  $start = "<TD VALIGN='TOP'><FONT COLOR='$col_tableheaderfg'>";
                  $end = "</FONT></TD>";
                 }
                 else
                 {
                  $start = "<TD VALIGN='TOP'>";
                  $end = "</TD>";
                 }
                }
                else
                {
                 $start = "[".$this->text."]";
                 $end = "[/td]";
                }
                break;
     case "table":

                $istable = false;
                $istable = $this->validateTable($this);
                $this->suppress = !$istable;
                
                if($istable)
                {
                 $this->ctr = 0;
                 $border = $this->paramList["border"];

                 if(is_numeric($border) && $border>0 && $border<=50)
                 {
                  $start = "<TABLE BORDER='0' CELLPADDING='$border' CELLSPACING='0'><TR BGCOLOR='$col_tableframe'><TD><TABLE BORDER='0' CELLPADDING='1' CELLSPACING='0'>";
                  $end = "</TABLE></TD></TR></TABLE>";
                 }
                 else
                 {
                  $start = "<TABLE BORDER='0' CELLPADDING='0' CELLSPACING='0'>";
                  $end = "</TABLE>";
                 }
                 $dotext = false;
                }
                else
                {
                 $start = "[".$this->text."]";
                 $end = "[/table]";
                }
                break;
     case "ul":
                $isok = false;
                $children = $this->children;

                // check if there are any [li] subnodes
                foreach($children AS $id => $child)
                {
                 if($child->type=="NODE" && $child->name=="li")
                 {
                  $isok = true;
                 }
                }

                if($isok)
                {
                 $dochildren = false;

                 $result.="<UL>";
                 foreach($children AS $id => $child)
                 {
                  // ignore anything that isn't a [li] node
                  if($child->type=="NODE" && $child->name=="li")
                  {
                   $result.=$child->render();
                  }
                 }
                 $result.="</UL>";

                 $start = "";
                 $end = "";
                }
                else
                {
                 $start = "[".$this->text."]";
                 $end = "[/ul]";
                }
                break;
     case "li":
                if($this->parent->name=="ul")
                {
                 $start = "<LI>";
                 $end = "</LI>";
                }
                else
                {
                 $start = "[".$this->text."]";
                 $end = "[/li]";
                }
                break;
     default:
                break;
    }

    $result.=$start;
    if($dochildren)
    {
     $children = $this->children;
     foreach($children AS $id => $child)
     {
      if($dotext)
      {
       $result.= $child->render();
      }
      else if($child->type=="NODE")
      {
       $result.= $child->render();
      }
     }
    }
    $result.=$end;
   }


   return $result;
  }

  function validateTable($table)
  {
   if($table->type!="NODE" || $table->name!="table")
   {
    return false;
   }


   $cellcount = array();
   $rowcount = 0;

   $rows = $table->children;
   foreach($rows AS $id => $row)
   {
    if($row->type=="NODE" && $row->name=="tr")
    {
     $cellcount[$rowcount] = 0;
     $cells = $row->children;
     foreach($cells AS $id => $cell)
     {
      if($cell->type=="NODE" && $cell->name=="td")
      {
       $cellcount[$rowcount] = $cellcount[$rowcount] + 1;
      }
     }
     $rowcount++;
    }
   }

   if($rowcount==0)
   {
    return false;
   }


   $cells = $cellcount[0];
   foreach($cellcount AS $id => $count)
   {
    if($count==0 || $cells!=$count)
    {
     return false;
    }
   }

   return true;
  }

  function extractParams()
  {
   $result = array();

   // must handle [url=xyz] differently to allow spaces in urls
   // even if it isn't valid html
   if($this->name=="url")
   {
    $pos = strpos($this->params,"=");
    if($pos!==false)
    {
     $value = substr($this->params,$pos+1);
     $result["url"] = $value;
    }
   }
   else // Handle all other tags normally
   {
    $data = explode(" ",$this->params);
    for($i=0;$i<count($data);$i++)
    {
     $pos = strpos($data[$i],"=");
     if($pos!==false)
     {
      $name = substr($data[$i],0,$pos);
      $value = substr($data[$i],$pos+1);
      if($name!="" && $value!="")
      {
       $result[$name] = $value;
      }
     }
    }
   }
   $this->paramList = $result;
  }

  function toString()
  {
   if($this->type=="TEXT")
   {
    return $this->text;
   }
   else
   {
    $out = $this->name;
    if(is_array($this->paramList) && count($this->paramList)!=0)
    {
     $out.= " params: [";
     foreach($this->paramList AS $name => $value)
     {
      $out.= " $name=$value";
     }
     $out.= "]";
    }

    return $out;
   }
  }
 }


?>
